package ru.dbotthepony.kstarbound.lua

import it.unimi.dsi.fastutil.ints.IntAVLTreeSet
import ru.dbotthepony.kstarbound.lua.userdata.LuaFuture
import ru.dbotthepony.kstarbound.lua.userdata.LuaPathFinder
import ru.dbotthepony.kstarbound.util.random.random
import java.io.Closeable
import java.lang.ref.Cleaner.Cleanable
import java.util.concurrent.ConcurrentLinkedQueue
import java.util.random.RandomGenerator
import kotlin.properties.Delegates

class LuaSharedState(val handlesThread: LuaThread, private val cleanable: Cleanable) : Closeable {
	private val pendingFree = ConcurrentLinkedQueue<Int>()
	private val freeHandles = IntAVLTreeSet()
	private var nextHandle = 0
	// faster code path
	private var handlesInUse = 0

	private val namedHandles = HashMap<Any, LuaHandle>()
	var random: RandomGenerator = random()

	var commonHandles by Delegates.notNull<CommonHandleRegistry>()
		private set

	var isValid = true
		private set

	fun ensureValid() {
		check(isValid) { "Tried to use NULL LuaState!" }
	}

	override fun close() {
		if (!isValid) return
		isValid = false
		namedHandles.clear()
		cleanable.clean()
	}

	fun initializeHandles(mainThread: LuaThread) {
		val future = LuaFuture.initializeHandle(mainThread)
		val pathFinder = LuaPathFinder.initializeHandle(mainThread)

		commonHandles = CommonHandleRegistry(
			future = future,
			pathFinder = pathFinder,
		)
	}

	fun freeHandle(handle: Int, key: Any?) {
		if (!isValid) return
		pendingFree.add(handle)

		if (key != null) {
			namedHandles.remove(key)
		}
	}

	fun cleanup() {
		ensureValid()
		if (handlesInUse == 0) return
		var handle = pendingFree.poll()

		while (handle != null) {
			handlesInUse--
			freeHandles.add(handle)
			handlesThread.push()
			handlesThread.copy(-1, handle)
			handlesThread.pop()

			handle = pendingFree.poll()
		}
	}

	fun allocateHandle(name: Any?): LuaHandle {
		ensureValid()
		require(name == null || name !in namedHandles) { "Named handle '$name' already exists" }
		handlesInUse++

		if (freeHandles.isEmpty()) {
			if (nextHandle % 10 == 0) {
				handlesThread.ensureExtraCapacity(10)
			}

			return LuaHandle(this, ++nextHandle, name).also {
				if (name != null) namedHandles[name] = it
			}
		} else {
			val handle = freeHandles.firstInt()
			freeHandles.remove(handle)

			handlesThread.copy(-1, handle)
			handlesThread.pop()

			return LuaHandle(this, handle, name).also {
				if (name != null) namedHandles[name] = it
			}
		}
	}

	fun getNamedHandle(key: Any): LuaHandle {
		ensureValid()
		return namedHandles[key] ?: throw NoSuchElementException("No such handle: $key")
	}

	fun findNamedHandle(key: Any): LuaHandle? {
		ensureValid()
		return namedHandles[key]
	}
}
